@@ -11,6 +11,7 @@
  * GNU General Public License for more details:
  *
  * Copyright (C) 2012 Google, Inc.
+ * Copyright (C) 2015 Azimut Electronics
  */
 
 #include <config.h>
@@ -37,38 +38,208 @@ G_DEFINE_TYPE (MMBearerQmi, mm_bearer_qmi, MM_TYPE_BASE_BEARER)
 
 #define GLOBAL_PACKET_DATA_HANDLE 0xFFFFFFFF
 
-enum {
-    PROP_0,
-    PROP_FORCE_DHCP,
-    PROP_LAST
-};
-
-static GParamSpec *properties[PROP_LAST];
-
 struct _MMBearerQmiPrivate {
     /* State kept while connected */
     QmiClientWds *client_ipv4;
+    guint packet_service_status_ipv4_indication_id;
+    guint event_report_ipv4_indication_id;
+
     QmiClientWds *client_ipv6;
+    guint packet_service_status_ipv6_indication_id;
+    guint event_report_ipv6_indication_id;
+
     MMPort *data;
     guint32 packet_data_handle_ipv4;
     guint32 packet_data_handle_ipv6;
-    gboolean force_dhcp;
 };
 
+/*****************************************************************************/
+/* Stats */
+
+typedef enum {
+    RELOAD_STATS_CONTEXT_STEP_FIRST,
+    RELOAD_STATS_CONTEXT_STEP_IPV4,
+    RELOAD_STATS_CONTEXT_STEP_IPV6,
+    RELOAD_STATS_CONTEXT_STEP_LAST,
+} ReloadStatsContextStep;
+
+typedef struct {
+    guint64 rx_bytes;
+    guint64 tx_bytes;
+} ReloadStatsResult;
+
+typedef struct {
+    MMBearerQmi *self;
+    GSimpleAsyncResult *result;
+    QmiMessageWdsGetPacketStatisticsInput *input;
+    ReloadStatsContextStep step;
+    ReloadStatsResult stats;
+} ReloadStatsContext;
+
+static gboolean
+reload_stats_finish (MMBaseBearer *bearer,
+                     guint64 *rx_bytes,
+                     guint64 *tx_bytes,
+                     GAsyncResult *res,
+                     GError **error)
+{
+    ReloadStatsResult *stats;
+
+    if (g_simple_async_result_propagate_error (G_SIMPLE_ASYNC_RESULT (res), error))
+        return FALSE;
+
+    stats = g_simple_async_result_get_op_res_gpointer (G_SIMPLE_ASYNC_RESULT (res));
+    if (rx_bytes)
+        *rx_bytes = stats->rx_bytes;
+    if (tx_bytes)
+        *tx_bytes = stats->tx_bytes;
+    return TRUE;
+}
+
+static void
+reload_stats_context_complete_and_free (ReloadStatsContext *ctx)
+{
+    g_simple_async_result_complete (ctx->result);
+    g_object_unref (ctx->result);
+    g_object_unref (ctx->self);
+    qmi_message_wds_get_packet_statistics_input_unref (ctx->input);
+    g_slice_free (ReloadStatsContext, ctx);
+}
+
+static void reload_stats_context_step (ReloadStatsContext *ctx);
+
+static void
+get_packet_statistics_ready (QmiClientWds *client,
+                             GAsyncResult *res,
+                             ReloadStatsContext *ctx)
+{
+    GError *error = NULL;
+    QmiMessageWdsGetPacketStatisticsOutput *output;
+    guint64 tx_bytes_ok = 0;
+    guint64 rx_bytes_ok = 0;
+
+    output = qmi_client_wds_get_packet_statistics_finish (client, res, &error);
+    if (!output) {
+        g_prefix_error (&error, "QMI operation failed: ");
+        g_simple_async_result_take_error (ctx->result, error);
+        reload_stats_context_complete_and_free (ctx);
+        return;
+    }
+
+    if (!qmi_message_wds_get_packet_statistics_output_get_result (output, &error)) {
+        g_prefix_error (&error, "Couldn't get packet statistics: ");
+        g_simple_async_result_take_error (ctx->result, error);
+        qmi_message_wds_get_packet_statistics_output_unref (output);
+        reload_stats_context_complete_and_free (ctx);
+        return;
+    }
+
+    qmi_message_wds_get_packet_statistics_output_get_tx_bytes_ok (output, &tx_bytes_ok, NULL);
+    qmi_message_wds_get_packet_statistics_output_get_rx_bytes_ok (output, &rx_bytes_ok, NULL);
+    ctx->stats.rx_bytes += rx_bytes_ok;
+    ctx->stats.tx_bytes += tx_bytes_ok;
+
+    qmi_message_wds_get_packet_statistics_output_unref (output);
+
+    /* Go on */
+    ctx->step++;
+    reload_stats_context_step (ctx);
+}
+
+static void
+reload_stats_context_step (ReloadStatsContext *ctx)
+{
+    switch (ctx->step) {
+    case RELOAD_STATS_CONTEXT_STEP_FIRST:
+        /* Fall through */
+        ctx->step++;
+    case RELOAD_STATS_CONTEXT_STEP_IPV4:
+        if (ctx->self->priv->client_ipv4) {
+            qmi_client_wds_get_packet_statistics (QMI_CLIENT_WDS (ctx->self->priv->client_ipv4),
+                                                  ctx->input,
+                                                  10,
+                                                  NULL,
+                                                  (GAsyncReadyCallback)get_packet_statistics_ready,
+                                                  ctx);
+            return;
+        }
+        ctx->step++;
+        /* Fall through */
+    case RELOAD_STATS_CONTEXT_STEP_IPV6:
+        if (ctx->self->priv->client_ipv6) {
+            qmi_client_wds_get_packet_statistics (QMI_CLIENT_WDS (ctx->self->priv->client_ipv6),
+                                                  ctx->input,
+                                                  10,
+                                                  NULL,
+                                                  (GAsyncReadyCallback)get_packet_statistics_ready,
+                                                  ctx);
+            return;
+        }
+        ctx->step++;
+        /* Fall through */
+    case RELOAD_STATS_CONTEXT_STEP_LAST:
+        g_simple_async_result_set_op_res_gpointer (ctx->result, &ctx->stats, NULL);
+        reload_stats_context_complete_and_free (ctx);
+        return;
+    }
+}
+
+static void
+reload_stats (MMBaseBearer *self,
+              GAsyncReadyCallback callback,
+              gpointer user_data)
+{
+    ReloadStatsContext *ctx;
+
+    ctx = g_slice_new0 (ReloadStatsContext);
+    ctx->self = g_object_ref (self);
+    ctx->result = g_simple_async_result_new (G_OBJECT (self),
+                                             callback,
+                                             user_data,
+                                             reload_stats);
+    ctx->input = qmi_message_wds_get_packet_statistics_input_new ();
+    qmi_message_wds_get_packet_statistics_input_set_mask (
+        ctx->input,
+        (QMI_WDS_PACKET_STATISTICS_MASK_FLAG_TX_BYTES_OK |
+         QMI_WDS_PACKET_STATISTICS_MASK_FLAG_RX_BYTES_OK),
+        NULL);
+    ctx->step = RELOAD_STATS_CONTEXT_STEP_FIRST;
+
+    reload_stats_context_step (ctx);
+}
+
 /*****************************************************************************/
 /* Connect */
 
+static void common_setup_cleanup_packet_service_status_unsolicited_events (MMBearerQmi *self,
+                                                                           QmiClientWds *client,
+                                                                           gboolean enable,
+                                                                           guint *indication_id);
+
+static void setup_event_report_unsolicited_events (MMBearerQmi *self,
+                                                   QmiClientWds *client,
+                                                   GCancellable *cancellable,
+                                                   GAsyncReadyCallback callback,
+                                                   gpointer user_data);
+
+static void cleanup_event_report_unsolicited_events (MMBearerQmi *self,
+                                                     QmiClientWds *client,
+                                                     guint *indication_id);
+
 typedef enum {
     CONNECT_STEP_FIRST,
     CONNECT_STEP_OPEN_QMI_PORT,
+    CONNECT_STEP_IP_METHOD,
     CONNECT_STEP_IPV4,
     CONNECT_STEP_WDS_CLIENT_IPV4,
     CONNECT_STEP_IP_FAMILY_IPV4,
+    CONNECT_STEP_ENABLE_INDICATIONS_IPV4,
     CONNECT_STEP_START_NETWORK_IPV4,
     CONNECT_STEP_GET_CURRENT_SETTINGS_IPV4,
     CONNECT_STEP_IPV6,
     CONNECT_STEP_WDS_CLIENT_IPV6,
     CONNECT_STEP_IP_FAMILY_IPV6,
+    CONNECT_STEP_ENABLE_INDICATIONS_IPV6,
     CONNECT_STEP_START_NETWORK_IPV6,
     CONNECT_STEP_GET_CURRENT_SETTINGS_IPV6,
     CONNECT_STEP_LAST
@@ -88,9 +259,13 @@ typedef struct {
     gboolean no_ip_family_preference;
     gboolean default_ip_family_set;
 
+    MMBearerIpMethod ip_method;
+
     gboolean ipv4;
     gboolean running_ipv4;
     QmiClientWds *client_ipv4;
+    guint packet_service_status_ipv4_indication_id;
+    guint event_report_ipv4_indication_id;
     guint32 packet_data_handle_ipv4;
     MMBearerIpConfig *ipv4_config;
     GError *error_ipv4;
@@ -98,6 +273,8 @@ typedef struct {
     gboolean ipv6;
     gboolean running_ipv6;
     QmiClientWds *client_ipv6;
+    guint packet_service_status_ipv6_indication_id;
+    guint event_report_ipv6_indication_id;
     guint32 packet_data_handle_ipv6;
     MMBearerIpConfig *ipv6_config;
     GError *error_ipv6;
@@ -111,6 +288,30 @@ connect_context_complete_and_free (ConnectContext *ctx)
     g_free (ctx->apn);
     g_free (ctx->user);
     g_free (ctx->password);
+
+    if (ctx->packet_service_status_ipv4_indication_id) {
+        common_setup_cleanup_packet_service_status_unsolicited_events (ctx->self,
+                                                                       ctx->client_ipv4,
+                                                                       FALSE,
+                                                                       &ctx->packet_service_status_ipv4_indication_id);
+    }
+    if (ctx->event_report_ipv4_indication_id) {
+        cleanup_event_report_unsolicited_events (ctx->self,
+                                                 ctx->client_ipv4,
+                                                 &ctx->event_report_ipv4_indication_id);
+    }
+    if (ctx->packet_service_status_ipv6_indication_id) {
+        common_setup_cleanup_packet_service_status_unsolicited_events (ctx->self,
+                                                                       ctx->client_ipv6,
+                                                                       FALSE,
+                                                                       &ctx->packet_service_status_ipv6_indication_id);
+    }
+    if (ctx->event_report_ipv6_indication_id) {
+        cleanup_event_report_unsolicited_events (ctx->self,
+                                                 ctx->client_ipv6,
+                                                 &ctx->event_report_ipv6_indication_id);
+    }
+
     g_clear_error (&ctx->error_ipv4);
     g_clear_error (&ctx->error_ipv6);
     g_clear_object (&ctx->client_ipv4);
@@ -281,6 +482,7 @@ qmi_inet4_ntop (guint32 address, char *buf, const gsize buflen)
 
 static MMBearerIpConfig *
 get_ipv4_config (MMBearerQmi *self,
+                 MMBearerIpMethod ip_method,
                  QmiMessageWdsGetCurrentSettingsOutput *output,
                  guint32 mtu)
 {
@@ -312,11 +514,7 @@ get_ipv4_config (MMBearerQmi *self,
     mm_info ("QMI IPv4 Settings:");
 
     config = mm_bearer_ip_config_new ();
-    if (self->priv->force_dhcp)
-        /* Force DHCP, but still set the static IPv4 config details if we get them */
-        mm_bearer_ip_config_set_method (config, MM_BEARER_IP_METHOD_DHCP);
-    else
-        mm_bearer_ip_config_set_method (config, MM_BEARER_IP_METHOD_STATIC);
+    mm_bearer_ip_config_set_method (config, ip_method);
 
     /* IPv4 address */
     qmi_inet4_ntop (addr, buf, sizeof (buf));
@@ -386,6 +584,7 @@ qmi_inet6_ntop (GArray *array, char *buf, const gsize buflen)
 
 static MMBearerIpConfig *
 get_ipv6_config (MMBearerQmi *self,
+                 MMBearerIpMethod ip_method,
                  QmiMessageWdsGetCurrentSettingsOutput *output,
                  guint32 mtu)
 {
@@ -408,11 +607,7 @@ get_ipv6_config (MMBearerQmi *self,
     mm_info ("QMI IPv6 Settings:");
 
     config = mm_bearer_ip_config_new ();
-    if (self->priv->force_dhcp)
-        /* Force DHCP, but still set the static IPv6 config details if we get them */
-        mm_bearer_ip_config_set_method (config, MM_BEARER_IP_METHOD_DHCP);
-    else
-        mm_bearer_ip_config_set_method (config, MM_BEARER_IP_METHOD_STATIC);
+    mm_bearer_ip_config_set_method (config, ip_method);
 
     /* IPv6 address */
     qmi_inet6_ntop (array, buf, sizeof (buf));
@@ -499,9 +694,9 @@ get_current_settings_ready (QmiClientWds *client,
         }
 
         if (ip_family == QMI_WDS_IP_FAMILY_IPV4)
-            ctx->ipv4_config = get_ipv4_config (ctx->self, output, mtu);
+            ctx->ipv4_config = get_ipv4_config (ctx->self, ctx->ip_method, output, mtu);
         else if (ip_family == QMI_WDS_IP_FAMILY_IPV6)
-            ctx->ipv6_config = get_ipv6_config (ctx->self, output, mtu);
+            ctx->ipv6_config = get_ipv6_config (ctx->self, ctx->ip_method, output, mtu);
 
         /* Domain names */
         if (qmi_message_wds_get_current_settings_output_get_domain_name_list (output, &array, &error)) {
@@ -588,6 +783,202 @@ set_ip_family_ready (QmiClientWds *client,
     connect_context_step (ctx);
 }
 
+static void
+packet_service_status_indication_cb (QmiClientWds *client,
+                                     QmiIndicationWdsPacketServiceStatusOutput *output,
+                                     MMBearerQmi *self)
+{
+    QmiWdsConnectionStatus connection_status;
+
+    if (qmi_indication_wds_packet_service_status_output_get_connection_status (
+            output,
+            &connection_status,
+            NULL,
+            NULL)) {
+        MMBearerStatus bearer_status = mm_base_bearer_get_status (MM_BASE_BEARER (self));
+
+        if (connection_status == QMI_WDS_CONNECTION_STATUS_DISCONNECTED &&
+            bearer_status != MM_BEARER_STATUS_DISCONNECTED &&
+            bearer_status != MM_BEARER_STATUS_DISCONNECTING) {
+            QmiWdsCallEndReason cer;
+            QmiWdsVerboseCallEndReasonType verbose_cer_type;
+            gint16 verbose_cer_reason;
+
+            if (qmi_indication_wds_packet_service_status_output_get_call_end_reason (
+                    output,
+                    &cer,
+                    NULL))
+                mm_info ("bearer call end reason (%u): '%s'",
+                         cer,
+                         qmi_wds_call_end_reason_get_string (cer));
+
+            if (qmi_indication_wds_packet_service_status_output_get_verbose_call_end_reason (
+                    output,
+                    &verbose_cer_type,
+                    &verbose_cer_reason,
+                    NULL))
+                mm_info ("bearer verbose call end reason (%u,%d): [%s] %s",
+                         verbose_cer_type,
+                         verbose_cer_reason,
+                         qmi_wds_verbose_call_end_reason_type_get_string (verbose_cer_type),
+                         qmi_wds_verbose_call_end_reason_get_string (verbose_cer_type, verbose_cer_reason));
+
+            mm_base_bearer_report_connection_status (MM_BASE_BEARER (self), MM_BEARER_CONNECTION_STATUS_DISCONNECTED);
+        }
+    }
+}
+
+static void
+common_setup_cleanup_packet_service_status_unsolicited_events (MMBearerQmi *self,
+                                                               QmiClientWds *client,
+                                                               gboolean enable,
+                                                               guint *indication_id)
+{
+    if (!client)
+        return;
+
+    /* Connect/Disconnect "Packet Service Status" indications */
+    if (enable) {
+        g_assert (*indication_id == 0);
+        *indication_id =
+            g_signal_connect (client,
+                              "packet-service-status",
+                              G_CALLBACK (packet_service_status_indication_cb),
+                              self);
+    } else if (*indication_id != 0) {
+        g_signal_handler_disconnect (client, *indication_id);
+        *indication_id = 0;
+    }
+}
+
+static void
+event_report_indication_cb (QmiClientWds *client,
+                            QmiIndicationWdsEventReportOutput *output,
+                            MMBearerQmi *self)
+{
+    mm_dbg ("Got QMI WDS event report");
+}
+
+static guint
+connect_enable_indications_ready (QmiClientWds *client,
+                                  GAsyncResult *res,
+                                  MMBearerQmi *self,
+                                  GError **error)
+{
+    QmiMessageWdsSetEventReportOutput *output;
+
+    /* Don't care about the result */
+    output = qmi_client_wds_set_event_report_finish (client, res, error);
+    if (!output || !qmi_message_wds_set_event_report_output_get_result (output, error)) {
+        if (output)
+            qmi_message_wds_set_event_report_output_unref (output);
+        return 0;
+    }
+    qmi_message_wds_set_event_report_output_unref (output);
+
+    return g_signal_connect (client,
+                             "event-report",
+                             G_CALLBACK (event_report_indication_cb),
+                             self);
+}
+
+static void
+connect_enable_indications_ipv4_ready (QmiClientWds *client,
+                                       GAsyncResult *res,
+                                       ConnectContext *ctx)
+{
+    g_assert (ctx->event_report_ipv4_indication_id == 0);
+
+    ctx->event_report_ipv4_indication_id =
+        connect_enable_indications_ready (client, res, ctx->self, &ctx->error_ipv4);
+    if (!ctx->event_report_ipv4_indication_id) {
+        ctx->step = CONNECT_STEP_LAST;
+        return;
+    }
+
+    ctx->step++;
+    connect_context_step (ctx);
+}
+
+static void
+connect_enable_indications_ipv6_ready (QmiClientWds *client,
+                                       GAsyncResult *res,
+                                       ConnectContext *ctx)
+{
+    g_assert (ctx->event_report_ipv6_indication_id == 0);
+
+    ctx->event_report_ipv6_indication_id =
+        connect_enable_indications_ready (client, res, ctx->self, &ctx->error_ipv6);
+    if (!ctx->event_report_ipv6_indication_id) {
+        ctx->step = CONNECT_STEP_LAST;
+        return;
+    }
+
+    ctx->step++;
+    connect_context_step (ctx);
+}
+
+static QmiMessageWdsSetEventReportInput *
+event_report_input_new (gboolean enable)
+{
+    QmiMessageWdsSetEventReportInput *input;
+
+    input = qmi_message_wds_set_event_report_input_new ();
+    qmi_message_wds_set_event_report_input_set_extended_data_bearer_technology (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_limited_data_system_status (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_uplink_flow_control (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_data_systems (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_evdo_pm_change (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_preferred_data_system (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_data_call_status (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_current_data_bearer_technology (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_mip_status (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_dormancy_status (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_data_bearer_technology (input, enable, NULL);
+    qmi_message_wds_set_event_report_input_set_channel_rate (input, enable, NULL);
+
+    return input;
+}
+
+static void
+setup_event_report_unsolicited_events (MMBearerQmi *self,
+                                       QmiClientWds *client,
+                                       GCancellable *cancellable,
+                                       GAsyncReadyCallback callback,
+                                       gpointer user_data)
+{
+    QmiMessageWdsSetEventReportInput *input = event_report_input_new (TRUE);
+
+    qmi_client_wds_set_event_report (client,
+                                     input,
+                                     5,
+                                     cancellable,
+                                     callback,
+                                     user_data);
+    qmi_message_wds_set_event_report_input_unref (input);
+}
+
+static void
+cleanup_event_report_unsolicited_events (MMBearerQmi *self,
+                                         QmiClientWds *client,
+                                         guint *indication_id)
+{
+    QmiMessageWdsSetEventReportInput *input;
+
+    g_assert (*indication_id != 0);
+    g_signal_handler_disconnect (client, *indication_id);
+    *indication_id = 0;
+
+    input = event_report_input_new (FALSE);
+    qmi_client_wds_set_event_report (client,
+                                     input,
+                                     5,
+                                     NULL,
+                                     NULL,
+                                     NULL);
+    qmi_message_wds_set_event_report_input_unref (input);
+}
+
 static void
 qmi_port_allocate_client_ready (MMPortQmi *qmi,
                                 GAsyncResult *res,
@@ -670,6 +1061,21 @@ connect_context_step (ConnectContext *ctx)
         /* If already open, just fall down */
         ctx->step++;
 
+    case CONNECT_STEP_IP_METHOD:
+        /* Once the QMI port is open, we decide the IP method we're going
+         * to request. If the LLP is raw-ip, we force Static IP, because not
+         * all DHCP clients support the raw-ip interfaces; otherwise default
+         * to DHCP as always. */
+        if (mm_port_qmi_llp_is_raw_ip (ctx->qmi))
+            ctx->ip_method = MM_BEARER_IP_METHOD_STATIC;
+        else
+            ctx->ip_method = MM_BEARER_IP_METHOD_DHCP;
+
+        mm_dbg ("Defaulting to use %s IP method", mm_bearer_ip_method_get_string (ctx->ip_method));
+
+        /* Just fall down */
+        ctx->step++;
+
     case CONNECT_STEP_IPV4:
         /* If no IPv4 setup needed, jump to IPv6 */
         if (!ctx->ipv4) {
@@ -731,6 +1137,18 @@ connect_context_step (ConnectContext *ctx)
         /* Just fall down */
         ctx->step++;
 
+    case CONNECT_STEP_ENABLE_INDICATIONS_IPV4:
+        common_setup_cleanup_packet_service_status_unsolicited_events (ctx->self,
+                                                                       ctx->client_ipv4,
+                                                                       TRUE,
+                                                                       &ctx->packet_service_status_ipv4_indication_id);
+        setup_event_report_unsolicited_events (ctx->self,
+                                               ctx->client_ipv4,
+                                               ctx->cancellable,
+                                               (GAsyncReadyCallback) connect_enable_indications_ipv4_ready,
+                                               ctx);
+        return;
+
     case CONNECT_STEP_START_NETWORK_IPV4: {
         QmiMessageWdsStartNetworkInput *input;
 
@@ -820,6 +1238,18 @@ connect_context_step (ConnectContext *ctx)
         /* Just fall down */
         ctx->step++;
 
+    case CONNECT_STEP_ENABLE_INDICATIONS_IPV6:
+        common_setup_cleanup_packet_service_status_unsolicited_events (ctx->self,
+                                                                       ctx->client_ipv6,
+                                                                       TRUE,
+                                                                       &ctx->packet_service_status_ipv6_indication_id);
+        setup_event_report_unsolicited_events (ctx->self,
+                                               ctx->client_ipv6,
+                                               ctx->cancellable,
+                                               (GAsyncReadyCallback) connect_enable_indications_ipv6_ready,
+                                               ctx);
+        return;
+
     case CONNECT_STEP_START_NETWORK_IPV6: {
         QmiMessageWdsStartNetworkInput *input;
 
@@ -860,6 +1290,10 @@ connect_context_step (ConnectContext *ctx)
             g_assert (ctx->self->priv->client_ipv4 == NULL);
             if (ctx->packet_data_handle_ipv4) {
                 ctx->self->priv->packet_data_handle_ipv4 = ctx->packet_data_handle_ipv4;
+                ctx->self->priv->packet_service_status_ipv4_indication_id = ctx->packet_service_status_ipv4_indication_id;
+                ctx->packet_service_status_ipv4_indication_id = 0;
+                ctx->self->priv->event_report_ipv4_indication_id = ctx->event_report_ipv4_indication_id;
+                ctx->event_report_ipv4_indication_id = 0;
                 ctx->self->priv->client_ipv4 = g_object_ref (ctx->client_ipv4);
             }
 
@@ -867,6 +1301,10 @@ connect_context_step (ConnectContext *ctx)
             g_assert (ctx->self->priv->client_ipv6 == NULL);
             if (ctx->packet_data_handle_ipv6) {
                 ctx->self->priv->packet_data_handle_ipv6 = ctx->packet_data_handle_ipv6;
+                ctx->self->priv->packet_service_status_ipv6_indication_id = ctx->packet_service_status_ipv6_indication_id;
+                ctx->packet_service_status_ipv6_indication_id = 0;
+                ctx->self->priv->event_report_ipv6_indication_id = ctx->event_report_ipv6_indication_id;
+                ctx->event_report_ipv6_indication_id = 0;
                 ctx->self->priv->client_ipv6 = g_object_ref (ctx->client_ipv6);
             }
 
@@ -984,6 +1422,7 @@ _connect (MMBaseBearer *self,
     ctx->data = data;
     ctx->cancellable = g_object_ref (cancellable);
     ctx->step = CONNECT_STEP_FIRST;
+    ctx->ip_method = MM_BEARER_IP_METHOD_UNKNOWN;
     ctx->result = g_simple_async_result_new (G_OBJECT (self),
                                              callback,
                                              user_data,
@@ -1201,6 +1640,14 @@ disconnect_context_step (DisconnectContext *ctx)
         if (ctx->packet_data_handle_ipv4) {
             QmiMessageWdsStopNetworkInput *input;
 
+            common_setup_cleanup_packet_service_status_unsolicited_events (ctx->self,
+                                                                           ctx->client_ipv4,
+                                                                           FALSE,
+                                                                           &ctx->self->priv->packet_service_status_ipv4_indication_id);
+            cleanup_event_report_unsolicited_events (ctx->self,
+                                                     ctx->client_ipv4,
+                                                     &ctx->self->priv->event_report_ipv4_indication_id);
+
             input = qmi_message_wds_stop_network_input_new ();
             qmi_message_wds_stop_network_input_set_packet_data_handle (input, ctx->packet_data_handle_ipv4, NULL);
 
@@ -1222,6 +1669,14 @@ disconnect_context_step (DisconnectContext *ctx)
         if (ctx->packet_data_handle_ipv6) {
             QmiMessageWdsStopNetworkInput *input;
 
+            common_setup_cleanup_packet_service_status_unsolicited_events (ctx->self,
+                                                                           ctx->client_ipv6,
+                                                                           FALSE,
+                                                                           &ctx->self->priv->packet_service_status_ipv6_indication_id);
+            cleanup_event_report_unsolicited_events (ctx->self,
+                                                     ctx->client_ipv6,
+                                                     &ctx->self->priv->event_report_ipv6_indication_id);
+
             input = qmi_message_wds_stop_network_input_new ();
             qmi_message_wds_stop_network_input_set_packet_data_handle (input, ctx->packet_data_handle_ipv6, NULL);
 
@@ -1318,8 +1773,7 @@ report_connection_status (MMBaseBearer *self,
 
 MMBaseBearer *
 mm_bearer_qmi_new (MMBroadbandModemQmi *modem,
-                   MMBearerProperties *config,
-                   gboolean force_dhcp)
+                   MMBearerProperties  *config)
 {
     MMBaseBearer *bearer;
 
@@ -1329,7 +1783,6 @@ mm_bearer_qmi_new (MMBroadbandModemQmi *modem,
     bearer = g_object_new (MM_TYPE_BEARER_QMI,
                            MM_BASE_BEARER_MODEM, modem,
                            MM_BASE_BEARER_CONFIG, config,
-                           MM_BEARER_QMI_FORCE_DHCP, force_dhcp,
                            NULL);
 
     /* Only export valid bearers */
@@ -1347,47 +1800,34 @@ mm_bearer_qmi_init (MMBearerQmi *self)
                                               MMBearerQmiPrivate);
 }
 
-static void
-set_property (GObject *object,
-              guint prop_id,
-              const GValue *value,
-              GParamSpec *pspec)
-{
-    MMBearerQmi *self = MM_BEARER_QMI (object);
-
-    switch (prop_id) {
-    case PROP_FORCE_DHCP:
-        self->priv->force_dhcp = g_value_get_boolean (value);
-        break;
-    default:
-        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-        break;
-    }
-}
-
-static void
-get_property (GObject *object,
-              guint prop_id,
-              GValue *value,
-              GParamSpec *pspec)
-{
-    MMBearerQmi *self = MM_BEARER_QMI (object);
-
-    switch (prop_id) {
-    case PROP_FORCE_DHCP:
-        g_value_set_boolean (value, self->priv->force_dhcp);
-        break;
-    default:
-        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-        break;
-    }
-}
-
 static void
 dispose (GObject *object)
 {
     MMBearerQmi *self = MM_BEARER_QMI (object);
 
+    if (self->priv->packet_service_status_ipv4_indication_id) {
+        common_setup_cleanup_packet_service_status_unsolicited_events (self,
+                                                                       self->priv->client_ipv4,
+                                                                       FALSE,
+                                                                       &self->priv->packet_service_status_ipv4_indication_id);
+    }
+    if (self->priv->event_report_ipv4_indication_id) {
+        cleanup_event_report_unsolicited_events (self,
+                                                 self->priv->client_ipv4,
+                                                 &self->priv->event_report_ipv4_indication_id);
+    }
+    if (self->priv->packet_service_status_ipv6_indication_id) {
+        common_setup_cleanup_packet_service_status_unsolicited_events (self,
+                                                                       self->priv->client_ipv6,
+                                                                       FALSE,
+                                                                       &self->priv->packet_service_status_ipv6_indication_id);
+    }
+    if (self->priv->event_report_ipv6_indication_id) {
+        cleanup_event_report_unsolicited_events (self,
+                                                 self->priv->client_ipv6,
+                                                 &self->priv->event_report_ipv6_indication_id);
+    }
+
     g_clear_object (&self->priv->data);
     g_clear_object (&self->priv->client_ipv4);
     g_clear_object (&self->priv->client_ipv6);
@@ -1405,21 +1845,12 @@ mm_bearer_qmi_class_init (MMBearerQmiClass *klass)
 
     /* Virtual methods */
     object_class->dispose = dispose;
-    object_class->get_property = get_property;
-    object_class->set_property = set_property;
 
     base_bearer_class->connect = _connect;
     base_bearer_class->connect_finish = connect_finish;
     base_bearer_class->disconnect = disconnect;
     base_bearer_class->disconnect_finish = disconnect_finish;
     base_bearer_class->report_connection_status = report_connection_status;
-
-    /* Properties */
-    properties[PROP_FORCE_DHCP] =
-        g_param_spec_boolean (MM_BEARER_QMI_FORCE_DHCP,
-                              "Force DHCP",
-                              "Never setup static IP config, always DHCP.",
-                              FALSE,
-                              G_PARAM_READWRITE);
-    g_object_class_install_property (object_class, PROP_FORCE_DHCP, properties[PROP_FORCE_DHCP]);
+    base_bearer_class->reload_stats = reload_stats;
+    base_bearer_class->reload_stats_finish = reload_stats_finish;
 }
