@@ -20,6 +20,9 @@
 
 #include <ModemManager.h>
 
+#define _LIBMM_INSIDE_MM
+#include <libmm-glib.h>
+
 #include "glib-object.h"
 #include "mm-charsets.h"
 
@@ -48,6 +51,8 @@ gchar       *mm_strip_quotes (gchar *str);
 const gchar *mm_strip_tag    (const gchar *str,
                               const gchar *cmd);
 
+gchar **mm_split_string_groups (const gchar *str);
+
 guint mm_count_bits_set (gulong number);
 
 gchar *mm_create_device_identifier (guint vid,
@@ -64,6 +69,28 @@ guint mm_netmask_to_cidr (const gchar *netmask);
 GArray *mm_filter_current_bands (const GArray *supported_bands,
                                  const GArray *current_bands);
 
+gchar *mm_new_iso8601_time (guint year,
+                            guint month,
+                            guint day,
+                            guint hour,
+                            guint minute,
+                            guint second,
+                            gboolean have_offset,
+                            gint offset_minutes);
+
+GArray *mm_filter_supported_modes (const GArray *all,
+                                   const GArray *supported_combinations);
+
+GArray *mm_filter_supported_capabilities (MMModemCapability all,
+                                          const GArray *supported_combinations);
+
+/*****************************************************************************/
+/* VOICE specific helpers and utilities */
+/*****************************************************************************/
+GRegex *mm_voice_ring_regex_get (void);
+GRegex *mm_voice_cring_regex_get(void);
+GRegex *mm_voice_clip_regex_get (void);
+
 /*****************************************************************************/
 /* 3GPP specific helpers and utilities */
 /*****************************************************************************/
@@ -76,6 +103,9 @@ GRegex    *mm_3gpp_cusd_regex_get (void);
 GRegex    *mm_3gpp_cmti_regex_get (void);
 GRegex    *mm_3gpp_cds_regex_get (void);
 
+/* AT+WS46=? response parser: returns array of MMModemMode values */
+GArray *mm_3gpp_parse_ws46_test_response (const gchar  *response,
+                                          GError      **error);
 
 /* AT+COPS=? (network scan) response parser */
 typedef struct {
@@ -89,6 +119,16 @@ void mm_3gpp_network_info_list_free (GList *info_list);
 GList *mm_3gpp_parse_cops_test_response (const gchar *reply,
                                          GError **error);
 
+/* AT+CGDCONT=? (PDP context format) test parser */
+typedef struct {
+    guint min_cid;
+    guint max_cid;
+    MMBearerIpFamily pdp_type;
+} MM3gppPdpContextFormat;
+void mm_3gpp_pdp_context_format_list_free (GList *pdp_format_list);
+GList *mm_3gpp_parse_cgdcont_test_response (const gchar *reply,
+                                            GError **error);
+
 /* AT+CGDCONT? (PDP context query) response parser */
 typedef struct {
     guint cid;
@@ -106,6 +146,7 @@ gboolean mm_3gpp_parse_creg_response (GMatchInfo *info,
                                       gulong *out_ci,
                                       MMModemAccessTechnology *out_act,
                                       gboolean *out_cgreg,
+                                      gboolean *out_cereg,
                                       GError **error);
 
 /* AT+CMGF=? (SMS message format) response parser */
@@ -120,6 +161,16 @@ gboolean mm_3gpp_parse_cpms_test_response (const gchar *reply,
                                            GArray **mem2,
                                            GArray **mem3);
 
+/* AT+CPMS? (Current SMS storage) response parser */
+gboolean mm_3gpp_parse_cpms_query_response (const gchar *reply,
+                                            MMSmsStorage *mem1,
+                                            MMSmsStorage *mem2,
+                                            GError** error);
+gboolean mm_3gpp_get_cpms_storage_match (GMatchInfo *match_info,
+                                         const gchar *match_name,
+                                         MMSmsStorage *storage,
+                                         GError **error);
+
 /* AT+CSCS=? (Supported charsets) response parser */
 gboolean mm_3gpp_parse_cscs_test_response (const gchar *reply,
                                            MMModemCharset *out_charsets);
@@ -149,6 +200,46 @@ gint         mm_3gpp_cind_response_get_max       (MM3gppCindResponse *r);
 GByteArray *mm_3gpp_parse_cind_read_response (const gchar *reply,
                                               GError **error);
 
+/* AT+CMGL=4 (list sms parts) response parser */
+typedef struct {
+    gint index;
+    gint status;
+    gchar *pdu;
+} MM3gppPduInfo;
+void   mm_3gpp_pdu_info_free           (MM3gppPduInfo *info);
+void   mm_3gpp_pdu_info_list_free      (GList *info_list);
+GList *mm_3gpp_parse_pdu_cmgl_response (const gchar *str,
+                                        GError **error);
+
+/* AT+CMGR (Read message) response parser */
+MM3gppPduInfo *mm_3gpp_parse_cmgr_read_response (const gchar *reply,
+                                                 guint index,
+                                                 GError **error);
+
+
+/* AT+CRSM response parser */
+gboolean mm_3gpp_parse_crsm_response (const gchar *reply,
+                                      guint *sw1,
+                                      guint *sw2,
+                                      gchar **hex,
+                                      GError **error);
+
+/* +CESQ response parser */
+gboolean mm_3gpp_parse_cesq_response (const gchar  *response,
+                                      guint        *out_rxlev,
+                                      guint        *out_ber,
+                                      guint        *out_rscp,
+                                      guint        *out_ecn0,
+                                      guint        *out_rsrq,
+                                      guint        *out_rsrp,
+                                      GError      **error);
+
+gboolean mm_3gpp_cesq_response_to_signal_info (const gchar  *response,
+                                               MMSignal    **out_gsm,
+                                               MMSignal    **out_umts,
+                                               MMSignal    **out_lte,
+                                               GError      **error);
+
 /* Additional 3GPP-specific helpers */
 
 MMModem3gppFacility mm_3gpp_acronym_to_facility (const gchar *str);
@@ -167,6 +258,8 @@ gboolean mm_3gpp_parse_operator_id (const gchar *operator_id,
 const gchar      *mm_3gpp_get_pdp_type_from_ip_family (MMBearerIpFamily family);
 MMBearerIpFamily  mm_3gpp_get_ip_family_from_pdp_type (const gchar *pdp_type);
 
+char *mm_3gpp_parse_iccid (const char *raw_iccid, GError **error);
+
 /*****************************************************************************/
 /* CDMA specific helpers and utilities */
 /*****************************************************************************/
@@ -202,4 +295,15 @@ gint  mm_cdma_normalize_class (const gchar *orig_class);
 gchar mm_cdma_normalize_band  (const gchar *long_band,
                                gint *out_class);
 
+gboolean mm_parse_gsn (const char *gsn,
+                       gchar **out_imei,
+                       gchar **out_meid,
+                       gchar **out_esn);
+
+/* +CCLK response parser */
+gboolean mm_parse_cclk_response (const gchar *response,
+                                 gchar **iso8601p,
+                                 MMNetworkTimezone **tzp,
+                                 GError **error);
+
 #endif  /* MM_MODEM_HELPERS_H */
